@@ -16,11 +16,12 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#ident "$Id: pty.c,v 1.25 2002/09/11 06:10:46 nalin Exp $"
+#ident "$Id: pty.c,v 1.33 2002/10/17 16:17:40 nalin Exp $"
 #include "../config.h"
-#include <sys/ioctl.h>
-#include <sys/socket.h>
 #include <sys/types.h>
+#include <sys/ioctl.h>
+#include <sys/uio.h>
+#include <sys/socket.h>
 #include <sys/wait.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -45,6 +46,12 @@
 
 #include "../gnome-pty-helper/gnome-pty.h"
 
+#ifdef MSG_NOSIGNAL
+#define PTY_RECVMSG_FLAGS MSG_NOSIGNAL
+#else
+#define PTY_RECVMSG_FLAGS 0
+#endif
+
 #ifdef ENABLE_NLS
 #include <libintl.h>
 #define _(String) dgettext(PACKAGE, String)
@@ -57,12 +64,147 @@ static pid_t _vte_pty_helper_pid = -1;
 static int _vte_pty_helper_tunnel = -1;
 static GTree *_vte_pty_helper_map = NULL;
 
+/* Reset the handlers for all known signals to their defaults.  The parent
+ * (or one of the libraries it links to) may have changed one to be ignored. */
+static void
+_vte_pty_reset_signal_handlers(void)
+{
+	signal(SIGHUP,  SIG_DFL);
+	signal(SIGINT,  SIG_DFL);
+	signal(SIGILL,  SIG_DFL);
+	signal(SIGABRT, SIG_DFL);
+	signal(SIGFPE,  SIG_DFL);
+	signal(SIGKILL, SIG_DFL);
+	signal(SIGSEGV, SIG_DFL);
+	signal(SIGPIPE, SIG_DFL);
+	signal(SIGALRM, SIG_DFL);
+	signal(SIGTERM, SIG_DFL);
+	signal(SIGCHLD, SIG_DFL);
+	signal(SIGCONT, SIG_DFL);
+	signal(SIGSTOP, SIG_DFL);
+	signal(SIGTSTP, SIG_DFL);
+	signal(SIGTTIN, SIG_DFL);
+	signal(SIGTTOU, SIG_DFL);
+#ifdef SIGBUS
+	signal(SIGBUS,  SIG_DFL);
+#endif
+#ifdef SIGPOLL
+	signal(SIGPOLL, SIG_DFL);
+#endif
+#ifdef SIGPROF
+	signal(SIGPROF, SIG_DFL);
+#endif
+#ifdef SIGSYS
+	signal(SIGSYS,  SIG_DFL);
+#endif
+#ifdef SIGTRAP
+	signal(SIGTRAP, SIG_DFL);
+#endif
+#ifdef SIGURG
+	signal(SIGURG,  SIG_DFL);
+#endif
+#ifdef SIGVTALARM
+	signal(SIGVTALARM, SIG_DFL);
+#endif
+#ifdef SIGXCPU
+	signal(SIGXCPU, SIG_DFL);
+#endif
+#ifdef SIGXFSZ
+	signal(SIGXFSZ, SIG_DFL);
+#endif
+#ifdef SIGIOT
+	signal(SIGIOT,  SIG_DFL);
+#endif
+#ifdef SIGEMT
+	signal(SIGEMT,  SIG_DFL);
+#endif
+#ifdef SIGSTKFLT
+	signal(SIGSTKFLT, SIG_DFL);
+#endif
+#ifdef SIGIO
+	signal(SIGIO,   SIG_DFL);
+#endif
+#ifdef SIGCLD
+	signal(SIGCLD,  SIG_DFL);
+#endif
+#ifdef SIGPWR
+	signal(SIGPWR,  SIG_DFL);
+#endif
+#ifdef SIGINFO
+	signal(SIGINFO, SIG_DFL);
+#endif
+#ifdef SIGLOST
+	signal(SIGLOST, SIG_DFL);
+#endif
+#ifdef SIGWINCH
+	signal(SIGWINCH, SIG_DFL);
+#endif
+#ifdef SIGUNUSED
+	signal(SIGUNUSED, SIG_DFL);
+#endif
+}
+
+#ifdef HAVE_SOCKETPAIR
+static int
+_vte_pty_pipe_open(int *a, int *b)
+{
+	int p[2], ret = -1;
+#ifdef PF_UNIX
+#ifdef SOCK_STREAM
+	ret = socketpair(PF_UNIX, SOCK_STREAM, 0, p);
+#else
+#ifdef SOCK_DGRAM
+	ret = socketpair(PF_UNIX, SOCK_DGRAM, 0, p);
+#endif
+#endif
+	if (ret == 0) {
+		*a = p[0];
+		*b = p[1];
+		return 0;
+	}
+#endif
+	return ret;
+}
+#else
+static int
+_vte_pty_pipe_open(int *a, int *b)
+{
+	int p[2], ret = -1;
+
+	ret = pipe(p);
+
+	if (ret == 0) {
+		*a = p[0];
+		*b = p[1];
+	}
+	return ret;
+}
+#endif
+
+static int
+_vte_pty_pipe_open_bi(int *a, int *b, int *c, int *d)
+{
+	int ret;
+	ret = _vte_pty_pipe_open(a, b);
+	if (ret != 0) {
+		return ret;
+	}
+	ret = _vte_pty_pipe_open(c, d);
+	if (ret != 0) {
+		close(*a);
+		close(*b);
+	}
+	return ret;
+}
+
 /* Run the given command, using the given descriptor as the controlling
  * terminal. */
 static int
-_vte_pty_run_on_pty(int fd, char **env_add, const char *command, char **argv)
+_vte_pty_run_on_pty(int fd, int ready_reader, int ready_writer,
+		    char **env_add, const char *command, char **argv)
 {
 	int i;
+	char c;
 	char **args, *arg;
 	if (fd != STDIN_FILENO) {
 		dup2(fd, STDIN_FILENO);
@@ -115,6 +257,35 @@ _vte_pty_run_on_pty(int fd, char **env_add, const char *command, char **argv)
 #endif
 	}
 
+	/* Reset our signals -- our parent may have done any number of
+	 * weird things to them. */
+	_vte_pty_reset_signal_handlers();
+
+	/* Signal to the parent that we've finished setting things up by
+	 * sending an arbitrary byte over the status pipe and waiting for
+	 * a response.  This synchronization step ensures that the pty is
+	 * fully initialized before the parent process attempts to do anything
+	 * with it, and is required on systems where additional setup, beyond
+	 * merely opening the device, is required.  This is at least the case
+	 * on Solaris. */
+#ifdef VTE_DEBUG
+	if (_vte_debug_on(VTE_DEBUG_PTY)) {
+		fprintf(stderr, "Child sending child-ready.\n");
+	}
+#endif
+	write(ready_writer, &c, 1);
+	fsync(ready_writer);
+	read(ready_reader, &c, 1);
+#ifdef VTE_DEBUG
+	if (_vte_debug_on(VTE_DEBUG_PTY)) {
+		fprintf(stderr, "Child received parent-ready.\n");
+	}
+#endif
+	close(ready_writer);
+	if (ready_writer != ready_reader) {
+		close(ready_reader);
+	}
+
 	/* Outta here. */
 	if (argv != NULL) {
 		for (i = 0; (argv[i] != NULL); i++) ;
@@ -137,12 +308,23 @@ _vte_pty_run_on_pty(int fd, char **env_add, const char *command, char **argv)
 /* Open the named PTY slave, fork off a child (storing its PID in child),
  * and exec the named command in its own session as a process group leader */
 static int
-_vte_pty_fork_on_pty_name(const char *path, char **env_add,
-			  const char *command, char **argv, pid_t *child)
+_vte_pty_fork_on_pty_name(const char *path, int parent_fd, char **env_add,
+			  const char *command, char **argv,
+			  int columns, int rows, pid_t *child)
 {
 	int fd, i;
+	char c;
+	int ready_a[2], ready_b[2];
 	pid_t pid;
 
+	/* Open pipes for synchronizing between parent and child. */
+	if (_vte_pty_pipe_open_bi(&ready_a[0], &ready_a[1],
+				  &ready_b[0], &ready_b[1]) == -1) {
+		/* Error setting up pipes.  Bail. */
+		*child = -1;
+		return -1;
+	}
+
 	/* Start up a child. */
 	pid = fork();
 	if (pid == -1) {
@@ -151,19 +333,53 @@ _vte_pty_fork_on_pty_name(const char *path, char **env_add,
 		return -1;
 	}
 	if (pid != 0) {
-		/* Parent.  Close our connection to the slave and return the
-		 * new child's PID. */
+		/* Parent.  Close the child's ends of the pipes, do the ready
+		 * handshake, and return the child's PID. */
+		close(ready_b[0]);
+		close(ready_a[1]);
+
+		/* Wait for the child to be ready, set the window size, then
+		 * signal that we're ready.  We need to synchronize here to
+		 * avoid possible races when the child has to do more setup
+		 * of the terminal than just opening it. */
+#ifdef VTE_DEBUG
+		if (_vte_debug_on(VTE_DEBUG_PTY)) {
+			fprintf(stderr, "Parent ready, waiting for child.\n");
+		}
+#endif
+		read(ready_a[0], &c, 1);
+#ifdef VTE_DEBUG
+		if (_vte_debug_on(VTE_DEBUG_PTY)) {
+			fprintf(stderr, "Parent received child-ready.\n");
+		}
+#endif
+		vte_pty_set_size(parent_fd, columns, rows);
+#ifdef VTE_DEBUG
+		if (_vte_debug_on(VTE_DEBUG_PTY)) {
+			fprintf(stderr, "Parent sending parent-ready.\n");
+		}
+#endif
+		write(ready_b[1], &c, 1);
+		close(ready_a[0]);
+		close(ready_b[1]);
+
 		*child = pid;
 		return 0;
 	}
 
-	/* Child.  Start a new session and become process-group leader. */
+	/* Child.  Close the parent's ends of the pipes. */
+	close(ready_a[0]);
+	close(ready_b[1]);
+
+	/* Start a new session and become process-group leader. */
 	setsid();
 	setpgid(0, 0);
 
 	/* Close all descriptors. */
 	for (i = 0; i < sysconf(_SC_OPEN_MAX); i++) {
-		close(i);
+		if ((i != ready_b[0]) && (i != ready_a[1])) {
+			close(i);
+		}
 	}
 
 	/* Open the slave PTY, acquiring it as the controlling terminal for
@@ -172,19 +388,31 @@ _vte_pty_fork_on_pty_name(const char *path, char **env_add,
 	if (fd == -1) {
 		return -1;
 	}
-	return _vte_pty_run_on_pty(fd, env_add, command, argv);
+	return _vte_pty_run_on_pty(fd, ready_b[0], ready_a[1],
+				   env_add, command, argv);
 }
 
 /* Fork off a child (storing its PID in child), and exec the named command
  * in its own session as a process group leader using the given terminal. */
 static int
 _vte_pty_fork_on_pty_fd(int fd, char **env_add,
-			const char *command, char **argv, pid_t *child)
+			const char *command, char **argv,
+			int columns, int rows, pid_t *child)
 {
 	int i;
 	char *tty;
+	char c;
+	int ready_a[2], ready_b[2];
 	pid_t pid;
 
+	/* Open pipes for synchronizing between parent and child. */
+	if (_vte_pty_pipe_open_bi(&ready_a[0], &ready_a[1],
+				  &ready_b[0], &ready_b[1]) == -1) {
+		/* Error setting up pipes.  Bail. */
+		*child = -1;
+		return -1;
+	}
+
 	/* Start up a child. */
 	pid = fork();
 	if (pid == -1) {
@@ -193,23 +421,55 @@ _vte_pty_fork_on_pty_fd(int fd, char **env_add,
 		return -1;
 	}
 	if (pid != 0) {
-		/* Parent.  Close our connection to the slave and return the
-		 * new child's PID. */
+		/* Parent.  Close the child's ends of the pipes, do the ready
+		 * handshake, and return the child's PID. */
+		close(ready_b[0]);
+		close(ready_a[1]);
+
+		/* Wait for the child to be ready, set the window size, then
+		 * signal that we're ready.  We need to synchronize here to
+		 * avoid possible races when the child has to do more setup
+		 * of the terminal than just opening it. */
+#ifdef VTE_DEBUG
+		if (_vte_debug_on(VTE_DEBUG_PTY)) {
+			fprintf(stderr, "Parent ready.\n");
+		}
+#endif
+		read(ready_a[0], &c, 1);
+#ifdef VTE_DEBUG
+		if (_vte_debug_on(VTE_DEBUG_PTY)) {
+			fprintf(stderr, "Parent received child-ready.\n");
+		}
+#endif
+		vte_pty_set_size(fd, columns, rows);
+#ifdef VTE_DEBUG
+		if (_vte_debug_on(VTE_DEBUG_PTY)) {
+			fprintf(stderr, "Parent sending parent-ready.\n");
+		}
+#endif
+		write(ready_b[1], &c, 1);
+		close(ready_a[0]);
+		close(ready_b[1]);
+
 		*child = pid;
 		return 0;
 	}
 
+	/* Child.  CLose the parent's ends of the pipes. */
+	close(ready_a[0]);
+	close(ready_b[1]);
+
 	/* Save the name of the pty -- we'll need it later to acquire it as
 	 * our controlling terminal. */
 	tty = ttyname(fd);
 
-	/* Child.  Start a new session and become process-group leader. */
+	/* Start a new session and become process-group leader. */
 	setsid();
 	setpgid(0, 0);
 
 	/* Close all other descriptors. */
 	for (i = 0; i < sysconf(_SC_OPEN_MAX); i++) {
-		if (i != fd) {
+		if ((i != fd) && (i != ready_b[0]) && (i != ready_a[1])) {
 			close(i);
 		}
 	}
@@ -223,7 +483,8 @@ _vte_pty_fork_on_pty_fd(int fd, char **env_add,
 		}
 	}
 
-	return _vte_pty_run_on_pty(fd, env_add, command, argv);
+	return _vte_pty_run_on_pty(fd, ready_b[0], ready_a[1],
+				   env_add, command, argv);
 }
 
 /**
@@ -403,11 +664,10 @@ _vte_pty_open_unix98(pid_t *child, char **env_add,
 			close(fd);
 			fd = -1;
 		} else {
-			/* Set the window size. */
-			vte_pty_set_size(fd, columns, rows);
 			/* Start up a child process with the given command. */
-			if (_vte_pty_fork_on_pty_name(buf, env_add, command,
-						      argv, child) != 0) {
+			if (_vte_pty_fork_on_pty_name(buf, fd, env_add, command,
+						      argv, columns, rows,
+						      child) != 0) {
 				close(fd);
 				fd = -1;
 			}
@@ -417,26 +677,66 @@ _vte_pty_open_unix98(pid_t *child, char **env_add,
 	return fd;
 }
 
-static int
-_vte_pty_pipe_open(int *a, int *b)
+#ifdef HAVE_SENDMSG
+static void
+_vte_pty_read_ptypair(int tunnel, int *parentfd, int *childfd)
 {
-	int p[2], ret = -1;
-#ifdef PF_UNIX
-#ifdef SOCK_STREAM
-	ret = socketpair(PF_UNIX, SOCK_STREAM, 0, p);
-#else
-#ifdef SOCK_DGRAM
-	ret = socketpair(PF_UNIX, SOCK_DGRAM, 0, p);
-#endif
-#endif
-	if (ret == 0) {
-		*a = p[0];
-		*b = p[1];
-		return 0;
+	int i, ret;
+	char control[LINE_MAX], iobuf[LINE_MAX];
+	struct cmsghdr *cmsg;
+	struct msghdr msg;
+	struct iovec vec;
+
+	for (i = 0; i < 2; i++) {
+		vec.iov_base = iobuf;
+		vec.iov_len = sizeof(iobuf);
+		msg.msg_name = NULL;
+		msg.msg_namelen = 0;
+		msg.msg_iov = &vec;
+		msg.msg_iovlen = 1;
+		msg.msg_control = control;
+		msg.msg_controllen = sizeof(control);
+		ret = recvmsg(tunnel, &msg, PTY_RECVMSG_FLAGS);
+		if (ret == -1) {
+			return;
+		}
+		for (cmsg = CMSG_FIRSTHDR(&msg);
+		     cmsg != NULL;
+		     cmsg = CMSG_NXTHDR(&msg, cmsg)) {
+			if (cmsg->cmsg_type == SCM_RIGHTS) {
+				memcpy(&ret, CMSG_DATA(cmsg), sizeof(ret));
+				switch (i) {
+					case 0:
+						*parentfd = ret;
+						break;
+					case 1:
+						*childfd = ret;
+						break;
+					default:
+						g_assert_not_reached();
+						break;
+				}
+			}
+		}
 	}
-#endif
-	return ret;
 }
+#else
+#ifdef I_RECVFD
+static void
+_vte_pty_read_ptypair(int tunnel, int *parentfd, int *childfd)
+{
+	int i;
+	if (ioctl(tunnel, I_RECVFD, &i) == -1) {
+		return;
+	}
+	*parentfd = i;
+	if (ioctl(tunnel, I_RECVFD, &i) == -1) {
+		return;
+	}
+	*childfd = i;
+}
+#endif
+#endif
 
 static void
 _vte_pty_stop_helper(void)
@@ -495,7 +795,7 @@ _vte_pty_start_helper(void)
 		close(tunnel);
 		close(_vte_pty_helper_tunnel);
 		/* Exec our helper. */
-		execl(PKGLIBDIR "/gnome-pty-helper",
+		execl(LIBEXECDIR "/gnome-pty-helper",
 		      "gnome-pty-helper", NULL);
 		/* Bail. */
 		_exit(1);
@@ -506,71 +806,10 @@ _vte_pty_start_helper(void)
 	return TRUE;
 }
 
-#ifdef SCM_RIGHTS
-static void
-_vte_pty_read_ptypair(int tunnel, int *parentfd, int *childfd)
-{
-	int i, ret;
-	char control[LINE_MAX], iobuf[LINE_MAX];
-	struct cmsghdr *cmsg;
-	struct msghdr msg;
-	struct iovec vec;
-
-	for (i = 0; i < 2; i++) {
-		vec.iov_base = iobuf;
-		vec.iov_len = sizeof(iobuf);
-		msg.msg_name = NULL;
-		msg.msg_namelen = 0;
-		msg.msg_iov = &vec;
-		msg.msg_iovlen = 1;
-		msg.msg_control = control;
-		msg.msg_controllen = sizeof(control);
-		ret = recvmsg(tunnel, &msg, MSG_NOSIGNAL);
-		if (ret == -1) {
-			return;
-		}
-		for (cmsg = CMSG_FIRSTHDR(&msg);
-		     cmsg != NULL;
-		     cmsg = CMSG_NXTHDR(&msg, cmsg)) {
-			if (cmsg->cmsg_type == SCM_RIGHTS) {
-				memcpy(&ret, CMSG_DATA(cmsg), sizeof(ret));
-				switch (i) {
-					case 0:
-						*parentfd = ret;
-						break;
-					case 1:
-						*childfd = ret;
-						break;
-					default:
-						g_assert_not_reached();
-						break;
-				}
-			}
-		}
-	}
-}
-#else
-#ifdef I_RECVFD
-static void
-_vte_pty_read_ptypair(int tunnel, int *parentfd, int *childfd)
-{
-	int i;
-	if (ioctl(tunnel, I_RECVFD, &i) == -1) {
-		return;
-	}
-	*parentfd = i;
-	if (ioctl(tunnel, I_RECVFD, &i) == -1) {
-		return;
-	}
-	*childfd = i;
-}
-#endif
-#endif
-
 static int
-_vte_pty_open_old_school(pid_t *child, char **env_add,
-			 const char *command, char **argv,
-			 int columns, int rows, int op)
+_vte_pty_open_with_helper(pid_t *child, char **env_add,
+			  const char *command, char **argv,
+			  int columns, int rows, int op)
 {
 	GnomePtyOps ops;
 	int ret;
@@ -604,20 +843,28 @@ _vte_pty_open_old_school(pid_t *child, char **env_add,
 		/* Receive the master and slave ptys. */
 		_vte_pty_read_ptypair(_vte_pty_helper_tunnel,
 				      &parentfd, &childfd);
+
 		if ((parentfd == -1) || (childfd == -1)) {
 			close(parentfd);
 			close(childfd);
 			return -1;
 		}
+
+#ifdef VTE_DEBUG
+		if (_vte_debug_on(VTE_DEBUG_MISC) ||
+		    _vte_debug_on(VTE_DEBUG_PTY)) {
+			fprintf(stderr, "Got master pty %d and slave pty %d.\n",
+				parentfd, childfd);
+		}
+#endif
+
 		/* Add the parent and the tag to our map. */
 		g_tree_insert(_vte_pty_helper_map,
 			      GINT_TO_POINTER(parentfd),
 			      tag);
-		/* Set the window size. */
-		vte_pty_set_size(parentfd, columns, rows);
 		/* Start up a child process with the given command. */
 		if (_vte_pty_fork_on_pty_fd(childfd, env_add, command,
-					    argv, child) != 0) {
+					    argv, columns, rows, child) != 0) {
 			close(parentfd);
 			close(childfd);
 			return -1;
@@ -677,8 +924,8 @@ vte_pty_open_with_logging(pid_t *child, char **env_add,
 	g_assert(op >= 0);
 	g_assert(op < G_N_ELEMENTS(opmap));
 	if (ret == -1) {
-		ret = _vte_pty_open_old_school(child, env_add, command, argv,
-					       columns, rows, opmap[op]);
+		ret = _vte_pty_open_with_helper(child, env_add, command, argv,
+						columns, rows, opmap[op]);
 	}
 	if (ret == -1) {
 		ret = _vte_pty_open_unix98(child, env_add, command, argv,
